/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 * \file position.h
 */

#ifndef POSITION_H
#define POSITION_H

#include <string.h>

#include "move.h"
#include "init.h"
#include "bit.h"

#define Max_Ply 500	

#define WHITE_CASTLE 3
#define WHITE_KING_CASTLE 1
#define WHITE_QUEEN_CASTLE 2
#define BLACK_CASTLE 12
#define BLACK_KING_CASTLE 4
#define BLACK_QUEEN_CASTLE 8

#define Opening 0
#define Endgame 1

#define positionCopy(from,to) memcpy(to, from, sizeof(Position))

struct tHistory{
	guint64 key;
	guint 	capture;
	guint 	staticMoves;
	guint 	enPassant;
	guint 	castle;
	gboolean	inCheck;
	gint    material[2];
	gint	pvt[2];
	gboolean	extend;
	gint	phase;
	Move 	move;
};
typedef struct tHistory History;

struct tPosition{
	guint64 piece[14];
	guint64 allPieces;
	guint   type[64];
	guint   turn;
	guint   ply;
	guint64 key;
	guint	capture;
	guint   staticMoves;
	guint   enPassant;
	guint   castle;
	gboolean	inCheck;
	gint    material[2];
	gint	pvt[2];
	gboolean extend;
	gint phase;
	History history[Max_Ply];
};
typedef struct tPosition Position;

Position* positionNew();
void positionFree(Position* position);
void positionStartPosition(Position* position);
void positionPrint(const Position* position);
guint positionRepeat(const Position* position);
void positionSetFen(Position* position, const gchar fen[]);
gboolean positionIsInCheck(const Position* position);
void positionIsOk(const Position* position);

static inline guint64 positionAttackBishop(const Position* position, const guint square){
	return *(magicmoves_b_indices[square]+
			(((position->allPieces&magicmoves_b_mask[square])*magicmoves_b_magics[square])>>magicmoves_b_shift[square]));
}
static inline guint64 positionAttackBishopAll(const guint square){
	return *(magicmoves_b_indices[square]+
			(((0ull&magicmoves_b_mask[square])*magicmoves_b_magics[square])>>magicmoves_b_shift[square]));
}

static inline guint64 positionAttackRook(const Position* position, const guint square){
	return *(magicmoves_r_indices[square]+
			(((position->allPieces&magicmoves_r_mask[square])*magicmoves_r_magics[square])>>magicmoves_r_shift[square]));
}

static inline guint64 positionAttackRookAll(const guint square){
	return *(magicmoves_r_indices[square]+
			(((0ull&magicmoves_r_mask[square])*magicmoves_r_magics[square])>>magicmoves_r_shift[square]));
}

static inline guint64 BmagicNOMASK(const unsigned int square,const guint64 occupancy){
	return *(magicmoves_b_indices[square]+
			(((occupancy)*magicmoves_b_magics[square])>>magicmoves_b_shift[square]));
}

static inline guint64 RmagicNOMASK(const unsigned int square, const guint64 occupancy){
				return *(magicmoves_r_indices[square]+(((occupancy)*magicmoves_r_magics[square])>>magicmoves_r_shift[square]));
}

static inline guint64 positionAttackQueen(const Position* position, const guint square){
	return positionAttackBishop(position,square)|positionAttackRook(position,square);
}

static inline guint64 positionAttackQueenAll(const guint square){
	return positionAttackBishopAll(square)|positionAttackRookAll(square);
}

static inline guint64 QmagicNOMASK(const unsigned int square, const guint64 occupancy){
		return BmagicNOMASK(square,occupancy)|RmagicNOMASK(square,occupancy);
}

static inline gint positionMaterial(Position* position){
	return position->material[position->turn] - position->material[position->turn^1];
}

gint moveMaterial(const Position* position, const Move move);

static inline gboolean positionLastMoveDangerous(const Position* position){
    Move move = position->history[position->ply-1].move;
    gint type = moveGetType(move), pieceType,square;
    if(type > Normal && type != Castle && type != Null_Move) return TRUE;
    if(type == Normal){
        square = moveGetFrom(move);
        pieceType = position->type[square];
        if(pieceType == White_Pawn && rank(square) >= Rank_6) return TRUE;
        if(pieceType == Black_Pawn && rank(square) <= Rank_3) return TRUE;
    }
    return FALSE;
}

static inline gboolean positionEvalEG(const Position* position){
	Move move =position->history[position->ply-1].move;
	if(moveGetType(move) == Capture || position->type[moveGetTo(move)] <= Black_Pawn)
		return TRUE;
	return FALSE;
}

#endif
