/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <glib/gprintf.h>

#include "position.h"
#include "search.h"
#include "threadSearch.h"
#include "position.h"
#include "moveGenerator.h"
#include "evaluator.h"
#include "moveDo.h"
#include "uci.h"
#include "hash.h"
#include "bit.h"
#include "book.h"
#include "evaluator.h"
#include "egbb.h"

gboolean stopFlag = FALSE;
gboolean parallelStopFlag = FALSE;
gboolean ponderSearch = FALSE;
guint searchMultipv = 1;
guint64 searchNodes;
guint history[64][64];
static gint searchDepth;

const gint futilityValue1 = 200;
const gint futilityMargin1 = 200;
const gint futilityValue2 = 500;
const gint futilityMargin2 = 500;
const gint futilityValue3 = 900;
const gint futilityMargin3 = 500;

const gint HistoryFullDepthMoves = 4;
const gint HistoryReductionLimit = 3;

gint iterativeDepth;

void rootSearch(Position* position, RootMoves* moves, guint depth);
guint updateMultipv(RootMoves* moves, guint current, gint eval, guint depth);

void iterativeDeepening(const Position* position, const guint64 maxNodes, const guint maxDepth){
	guint depth	= 1,i,j;
	Position searchPosition;
	Move bMove = Empty;

	searchNodes = 0;
	searchDepth = 0;
	bbhits = 0;
	
	positionCopy(position,&searchPosition);

	if(maxDepth == 0){
		   evaluate(&searchPosition, -EVAL_MATE, EVAL_MATE,TRUE);
		   g_printf("bestmove xxxx\n");
		   fflush(stdout);
		   return;
	}

	RootMoves moves;
	rootMovesGenerate(&searchPosition, &moves);

	if(positionRepeat(&searchPosition) == 2 || searchPosition.staticMoves >= 100 || moves.size == 0){
		g_printf("info string in given position can not make any move!\n");
		return;
	}
	
	if(moves.size == 1){
		uciBestMove(moves.move[0].move, Empty);
		return;
	}
	
	hashIncrement();

	for(i=0; i<64; i++)
		for(j=0; j<64; j++)
			history[i][j] = 0;

	searchMultipv = MIN(searchMultipv,moves.size);

	if(isBookLoaded()){
		bMove = bookMove(&searchPosition);
		if(bMove != Empty){
			uciBestMove(bMove,Empty);
			return;
		}
	}

	while(depth <= maxDepth && searchNodes < maxNodes && (!stopFlag || ponderSearch)){
		uciBeginDepthInfo(depth);
		iterativeDepth = depth;
		rootSearch(&searchPosition,&moves, depth);
		uciEndDepthInfo(depth, searchNodes);
		depth++;
	}

	uciBestMove(moves.move[0].move,moves.move[0].pv[0]);
}

void rootSearch(Position* position, RootMoves* moves, guint depth){
	gint eval, alpha, beta;
	guint i, current = 0, window;
	gboolean up, down;
	HashData* hd;
	guint64 key = position->key, hData;
	gboolean nodePV;
	
	hashFind(&hd,key,&hData);
			
	while(!stopFlag || ponderSearch){
		for(i=0; i<moves->size; i++){
			if(moves->move[i].depth < depth){
				current = i;
				break;
			}
		}
		if(i == moves->size) break;
		
		if(!moveDo(position,moves->move[current].move)) { g_error("Wrong root move!\n"); }
		uciCurrentMove(depth,current,moves);
		
		if(current < searchMultipv){
			eval = moves->move[current].eval;
			alpha = eval - 1;
			nodePV = TRUE;
		}
		else{
			eval = moves->move[searchMultipv-1].eval;
 			alpha = eval;
			nodePV = FALSE;
		}

		window = 2;
		up = down = FALSE;
		beta = eval + 1;

		while(TRUE){
			if(searchThreads == 1) eval = -search(position,-beta, -alpha, depth-1, 0, TRUE, nodePV);
			else eval = -threadSearch(position, -beta, -alpha, depth-1, 0, TRUE, nodePV);
			if(stopFlag == TRUE && !ponderSearch) break;
			if(current < searchMultipv){
				current = updateMultipv(moves, current, eval, depth);
				hashCollectPV(position, moves->move[current].pv);
				uciMultiPv(position,depth,searchMultipv,moves);
				if(eval >= beta){
					up = TRUE;
					if(up && down) break;
					window += window;
					alpha = eval;
					beta = eval + window;
					if(beta > EVAL_MATE) beta = EVAL_MATE+1;
					hashWrite(hd,key,depth,eval,ALPHA,moves->move[current].move);	
					continue;
				}
				if(eval <= alpha){
					down = TRUE;
					if(up && down) break;
					window += window;
					beta = eval;
					alpha = eval - window;
					if(alpha < -EVAL_MATE) alpha = -EVAL_MATE - 1;
					hashWrite(hd,key,depth,eval,BETA,moves->move[current].move);
					continue;
				}
				hashWrite(hd,key,depth,eval,EXACT,moves->move[current].move);
				break;
			}
			else{
				if(eval <= alpha){
					moves->move[current].eval = eval;
					moves->move[current].depth = depth;
					hashWrite(hd,key,depth,eval,BETA,moves->move[current].move);
					break;
				}
				else{
					current = updateMultipv(moves, current, eval, depth);
					uciMultiPv(position,depth,searchMultipv,moves);
					up = TRUE;
					window = 4;
					alpha = eval;
					beta = eval + window;
					if(beta > EVAL_MATE) beta = EVAL_MATE+1;
					hashWrite(hd,key,depth,eval,ALPHA,moves->move[current].move);
					nodePV = TRUE;
					continue;
				}
			}
		}
		moveUndo(position);
	}
	hashWriteBestMove(position->key,depth,moves->move[0].eval,moves->move[0].move);
}

guint updateMultipv(RootMoves* moves, guint current, gint eval, guint depth){
	gdouble sec = uciGetTime();
	RootMove move;
	guint i,nps;

	if(sec > 0) nps = (guint)(searchNodes/sec);
	else nps = 0;

	moves->move[current].eval = eval;
	moves->move[current].depth = depth;
	moves->move[current].seldepth = searchDepth;
	moves->move[current].time = (guint)(sec*1000);
	moves->move[current].nps = nps;
	moves->move[current].nodes = searchNodes;
	for(i=0; i<current; i++){
		if(moves->move[i].eval < eval){
			memcpy(&move,&moves->move[current],sizeof(RootMove));
			while(i < current){
				memcpy(&moves->move[current],&moves->move[current-1],sizeof(RootMove));
				current --;
			}
			memcpy(&moves->move[i],&move,sizeof(RootMove));
			current = i;
			break;
		}
	}
	return current;
}

gint search(Position* position, gint alpha, gint beta, gint depth, const gint hight, const gboolean nullMove, const gboolean nodePV){
	Moves moves;
	Move move, bestMove = Empty, hMove = Empty;
	gboolean extend=FALSE, endGamePhase, fprune = FALSE;
	gint eval, tmpDepth, fmargin = 0, movesSearched = 0;
	guint64 key = position->key, hData;
	HashData* hd;

	if((stopFlag == TRUE && !ponderSearch) || parallelStopFlag == TRUE ) return EVAL_DRAW;

	searchNodes++;

	/* Draw */
	if(positionRepeat(position) >= 1 || position->staticMoves >= 100) return EVAL_DRAW;
	
	/* End game */
	gint numberOfPieces = numberOfBits64(position->allPieces);
	if(numberOfPieces <= 5 && egbbIsLoaded()){
		tmpDepth = (2 * iterativeDepth) / 3;
		if((numberOfPieces <= 4 || hight <= tmpDepth) &&
			(hight >= tmpDepth || positionEvalEG(position))){
			if(egbbProbe(position, &eval)){
				if(eval < 0){
					   return eval + hight * 40 + positionMaterial(position);
				}
				if(eval > 0){
					   return eval - hight * 40 + positionMaterial(position);
				}
				return positionMaterial(position)>>4;
			}
		}
/*		else{
			egbbProbe(position, &eval);
			if(eval == 0) return positionMaterial(position)>>4;
		}*/
	}


	/* Mate distance pruning */
	eval = -EVAL_MATE + hight;
	if(eval > alpha){
		if(eval >= beta) return eval;
		alpha = eval;
	}

	eval = EVAL_MATE - hight;
	if(eval < beta){
		if(eval <= alpha) return eval;
		beta = eval;
	}

	/* Hash */
	if(hashFind(&hd,key,&hData)){
		hMove = hashMove(hData);
        if(!nodePV && (gint)hashDepth(hData) >= depth){
			eval = hashEval(hData);
			if(hashFlag(hData) == EXACT) return eval;
			else if(hashFlag(hData) == ALPHA && eval > alpha){
				bestMove = hashMove(hData);
				alpha = eval;
                if(alpha >= beta) return alpha;
			}
			else if(hashFlag(hData) == BETA && eval < beta){
				beta = eval;
				if(alpha >= beta) return beta;
			}
		}
	}

	/* Extension */	
	if(position->extend || (nodePV && passedPawnMove(position))){
		depth ++;
		extend = TRUE;
	}


	/* TO DO
	if(position->inCheck)
		extend = TRUE;
	
	if(position->capture != Empty &&
		position->history[position->ply-1].capture == position->capture)
		extend = TRUE;

	move = position->history[position->ply-1].move;
	guint to = moveGetTo(move);
	if(position->type[to] <= Black_Pawn && (to >= a7 || to <= h2))
		extend = TRUE;

	if(moveGetType(move) >= Promote_Queen)
		extend = TRUE;

	if(extend && (egbbIsLoaded() && numberOfPieces > 5)) depth ++;
	 */

	/* Horizon */
	if(depth < 1) return quiescent(position,alpha,beta,hight+1,nodePV);

	if(position->phase <= 7 && numberOfPieces <= 10) endGamePhase = TRUE;
	else endGamePhase = FALSE;

	/* Null move pruning */
	if (!nodePV && nullMove && depth>2 && !extend && !endGamePhase){
		if(depth > 6) tmpDepth = depth - 4;
		else tmpDepth = depth - 3;
		moveDo(position,Null_Move);
		eval = -search(position,-beta,-beta+1,tmpDepth,hight+1,FALSE,FALSE);
		moveUndo(position);
		if(eval >= beta){
			hashWrite(hd,key,depth,eval,ALPHA,Empty);
			return eval;
		}
	}

	/* Internal Iterative Deepening */
	if(nodePV && hMove == Empty && depth > 3 && depth/2 >= (gint)hashDepth(hData)){
		search(position,alpha, beta, depth/2, hight,FALSE, TRUE);
		if(hashFind2(hd,key,&hData)) hMove = hashMove(hData);
	}


   /* TO DO: Razoring: idea by Tord Romstad (Glaurung)
   else if(node_type != NodePV && depth <= RazorDepth &&  eval(board,alpha, beta, ThreadId) < beta - RazorMargin){
		value = full_quiescence(board,alpha,beta,0,height,pv,ThreadId);
      if(value < beta)
        return value;
   }*/


	/* Futility pruning */
	if(!nodePV && depth <= 3 && !extend && !endGamePhase){
		eval = positionMaterial(position);
		/* TO DO
		 * eval = evaluate(position, alpha, beta, nodePV, extension);
		 */
		switch(depth){
			case 3:{
				if(eval + futilityValue3 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin3;
					depth --;
					break;
				}
			}
			case 2:{
				if(eval + futilityValue2 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin2;
					break;
				}
			}
			case 1:{
				if(eval + futilityValue1 <= alpha){
					fprune = TRUE;
					fmargin = futilityMargin1;
					break;
				}
			}
		}
	}

	/* All moves */
	moves.state = START;
	if(hMove != Empty){
		move = hMove;
		moveDo(position,move);
	}
	else move = nextMoveDo(position,&moves);
	while(move != Empty){
		/* Futility pruning */
		if(fprune && !position->extend && fmargin - positionMaterial(position) <= alpha){
			moveUndo(position);
			move = nextMoveDo(position,&moves);
			movesSearched ++;
			continue;
		}
		if(movesSearched == 0){
			eval = -search(position, -beta, -alpha, depth-1, hight+1,TRUE, nodePV);
		}else{
			/* History pruning */
			if(!nodePV && movesSearched >= HistoryFullDepthMoves && depth > HistoryReductionLimit && 
				!extend && !position->extend){
				eval = -search(position, -(alpha+1), -alpha, depth-2, hight+1,TRUE,FALSE);
			}
			else eval = alpha + 1;

			/* PVS */
			if(eval > alpha){
				eval = -search(position, -(alpha+1), -alpha, depth-1, hight+1,TRUE,FALSE);
				if(eval > alpha && eval < beta)
					eval = -search(position, -beta, -alpha, depth-1, hight+1,TRUE,nodePV);
			}
		}
		/* TO DO Extension (pawn candidate)*/
		moveUndo(position);
		if(eval > alpha){
			if(eval >= beta){
				if(moves.state == STATIC) history[moveGetFrom(move)][moveGetTo(move)] += 1<<depth;
				hashWrite(hd,key,depth,eval,ALPHA,move);
				return eval;
			}
			alpha = eval;
			bestMove = move;
		}
		movesSearched ++;
		move = nextMoveDo(position,&moves);
	}

	/* Mate or pat */
	if(movesSearched == 0){
		if(position->inCheck == TRUE){
			eval = -EVAL_MATE+hight;
			hashWrite(hd,key,depth,eval,EXACT,Empty);
			return eval;
		}
		else{  
			hashWrite(hd,key,depth,EVAL_DRAW,EXACT,Empty);
			return EVAL_DRAW;
		}
	}

	/* hash is update */
	if(bestMove != Empty) hashWrite(hd,key,depth,alpha,EXACT,bestMove);
	else hashWrite(hd,key,depth,alpha,BETA,hMove);

	return alpha;
}

gint quiescent(Position* position, gint alpha, gint beta, const gint hight, const gboolean nodePV){
	Moves moves;
	Move move;
	gint eval;	

	if((stopFlag == TRUE && !ponderSearch) || parallelStopFlag == TRUE) return -EVAL_MATE;

	searchNodes++;

	if(hight > searchDepth) searchDepth = hight;
	
	/* End game */
	gint numberOfPieces = numberOfBits64(position->allPieces);
	if(numberOfPieces <= 5 && egbbIsLoaded()){
		gint tmpDepth = (2 * iterativeDepth) / 3;
		if((numberOfPieces <= 4 || hight <= tmpDepth) &&
			(hight >= tmpDepth || positionEvalEG(position))){
			if(egbbProbe(position, &eval)){
				if(eval < 0) return eval + hight * 40 + positionMaterial(position);
				if(eval > 0) return eval - hight * 40 + positionMaterial(position);
				return eval;
			}
		}
	}

	/* Mate distance pruning */
	eval = -EVAL_MATE + hight;
	if(eval > alpha){
		if(eval >= beta) return eval;
		alpha = eval;
	}

	eval = EVAL_MATE - hight;
	if(eval < beta){
		if(eval <= alpha) return eval;
		beta = eval;
	}

	eval = evaluate(position, alpha, beta, nodePV);

	if(eval > alpha)
	{
		if(eval >= beta) return eval;
		alpha = eval;
	}

	moves.state = START;
	move = nextDinamicMoveDo(position,&moves);
	while(move != Empty){
		eval = -quiescent(position, -beta, -alpha, hight+1, nodePV);
		moveUndo(position);
		if(eval > alpha){
			if(eval >= beta) return eval;
			alpha = eval;
		}
		move = nextDinamicMoveDo(position,&moves);
	}
	return alpha;
}

gboolean passedPawnMove(const Position* position){
	gint to = moveGetTo(position->history[position->ply-1].move);
	gint piece = position->type[to]; 
	if(piece == White_Pawn){
		gint pawnFile = file(to);
		guint64 bits = getBitFile(position->piece[Black_Pawn], pawnFile);
		bits |= getBitFile(position->piece[Black_Pawn], pawnFile - 1);
		bits |= getBitFile(position->piece[Black_Pawn], pawnFile + 1);
		if(!bitsGT(bits,rank(to))) return TRUE;
	}
	if(piece == Black_Pawn){
		gint pawnFile = file(to);
		guint64 bits = getBitFile(position->piece[White_Pawn], pawnFile);
		bits |= getBitFile(position->piece[White_Pawn], pawnFile - 1);
		bits |= getBitFile(position->piece[White_Pawn], pawnFile + 1);
		if(!bitsLE(bits,rank(to))) return TRUE;
	}
	return FALSE;
}
